SET SERVEROUTPUT ON;

/*Auto value column with identity*/
/*this can only be used for one column though*/
DROP TABLE books;

CREATE TABLE books (
  book_id NUMBER GENERATED BY DEFAULT ON NULL AS IDENTITY,
  title VARCHAR2(100) NOT NULL,
  created TIMESTAMP DEFAULT SYSDATE
);

/*since book_id is now automatically generated, we don't need to include it in */
/*the INSERT statement anymore*/
INSERT INTO books (title) VALUES ('book 1');
INSERT INTO books (title) VALUES ('book 2');
INSERT INTO books (title) VALUES ('book 3');

SELECT * FROM books;

/******/
/*simple trigger that insert a message to a log table
when a new book is inserted*/
DROP TABLE books;

CREATE TABLE books (
  book_id NUMBER GENERATED BY DEFAULT ON NULL AS IDENTITY, 
  title VARCHAR2(100) NOT NULL,
  created TIMESTAMP DEFAULT SYSDATE
);

CREATE TABLE book_logs (message VARCHAR2(100));

CREATE OR REPLACE TRIGGER insert_book_log
BEFORE INSERT ON books
FOR EACH ROW
BEGIN
    INSERT INTO book_logs VALUES('new book inserted by ' || USER || ' on ' || SYSDATE);
END;

INSERT INTO books (title) VALUES ('book 1');
INSERT INTO books (title) VALUES ('book 2');
INSERT INTO books (title) VALUES ('book 3');

SELECT * FROM books;
SELECT * FROM book_logs;

/******/
/*auto value with sequence and trigger*/
DROP TABLE books;

CREATE TABLE books (
  book_id INT PRIMARY KEY,
  title VARCHAR2(100) NOT NULL,
  created TIMESTAMP DEFAULT SYSDATE
);

/*this time we use a sequence to generate auto number*/
/*oracle sequence is an object that is used to generate number*/
/*we can set start point, and increment step*/
DROP SEQUENCE book_seq;

CREATE SEQUENCE book_seq
    MINVALUE 100
    START WITH 100
    INCREMENT BY 1;

/*this trigger will be activated BEFORE a row is inserted into the books table*/
/*here, :new represents the set of values that will be used to inserted as a new row to books*/
/*so, :new.book_id means the id of the new row*/
CREATE OR REPLACE TRIGGER insert_books
BEFORE INSERT ON books
FOR EACH ROW
BEGIN
    SELECT book_seq.NEXTVAL 
    INTO :new.book_id
    FROM DUAL;
END;

INSERT INTO books (title) VALUES ('book 1');
INSERT INTO books (title) VALUES ('book 2');
INSERT INTO books (title) VALUES ('book 3');

SELECT * FROM books;


/*****/
/*More examples with trigger*/

/*create table accounts*/
DROP TABLE account_infos;
DROP TABLE accounts;

/*similar like before, table accounts has aid as an auto-increment column*/
/*created and modified are the timestamp when the row is inserted*/
/*modified will changed when the row is updated, we can use a trigger to do this*/
CREATE TABLE accounts(
    aid INT GENERATED BY DEFAULT ON NULL AS IDENTITY,
    username VARCHAR(50) NOT NULL UNIQUE,
    profile_picture VARCHAR(100) NULL,
    created TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    modified TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT accountpk PRIMARY KEY (aid)
);

INSERT INTO accounts (username) VALUES ('user 1');
INSERT INTO accounts (username) VALUES ('user 2');
INSERT INTO accounts (username) VALUES ('user 3');

SELECT * FROM accounts;

/*in this example, we use a trigger to prevent updating username*/
/*if update other fields, we also update modified to the update time*/
CREATE OR REPLACE TRIGGER block_username_update
BEFORE UPDATE ON accounts
FOR EACH ROW
BEGIN
    IF UPDATING('username') THEN
        raise_application_error (-20999, 'cannot change username');
    ELSE
        :new.modified := SYSDATE;
        dbms_output.put_line('OK');
  END IF;
END;

SHOW ERRORS;

UPDATE accounts SET profile_picture = 'path to pic 1' WHERE aid = 1;
UPDATE accounts SET username = 'another username' WHERE aid = 1;

SELECT * FROM accounts;


/*Trigger to enforce 1-1 relationship*/
/*we create table account_infos with more personal details of an account*/
/*assume account_infos and accounts have an 1-1 relationship*/
/*we will use a trigger to make sure when one row is inserted into accounts*/
/*a row with the same aid is inserted into account_infos*/

CREATE TABLE account_infos (
    aid INT NOT NULL PRIMARY KEY,
    first_name VARCHAR(50) NULL,
    last_name VARCHAR(50) NULL,
    email VARCHAR(50) NULL,
    CONSTRAINT acc_info_fk FOREIGN KEY (aid) REFERENCES accounts (aid)
);

INSERT INTO account_infos VALUES (1,'Alice','A','alice@mail.com');
INSERT INTO account_infos VALUES (2,'Bob','B','bob@mail.com');
INSERT INTO account_infos VALUES (3,'Chad','C','chad@mail.com');

SELECT * FROM account_infos;

/*this trigger will insert a new row into account_infos*/
/*whenever a new row is inserted into accounts*/
/*we use the same aid of the new accounts row, and all other columns*/
/*in the account_infos row are left as NULL*/
CREATE OR REPLACE TRIGGER insert_accounts
AFTER INSERT ON accounts
FOR EACH ROW
    BEGIN
        INSERT INTO account_infos VALUES (:new.aid, '', '', '');
    END;
    
INSERT INTO accounts (username) VALUES ('user 4');
INSERT INTO accounts (username) VALUES ('user 5');

SELECT * FROM accounts;
SELECT * FROM account_infos;


/*Trigger to insert into joined views*/
/*this is a different approach from the previous example*/
/*in this case, we create a joined view for accounts and account_infos*/
/*and use it to insert new rows into the two tables*/
/*the previous trigger and the one in this example cannot be active at the same time*/
CREATE OR REPLACE VIEW account_details AS
SELECT ac.aid, ac.username, ad.first_name, ad.last_name, ad.email
FROM accounts ac
INNER JOIN account_infos ad
ON ac.aid = ad.aid;

SELECT * FROM account_details;

/*we need to disable the previous trigger 
since it will overlap this one*/
ALTER TRIGGER insert_accounts DISABLE;

CREATE OR REPLACE TRIGGER insert_account_details
INSTEAD OF INSERT ON account_details
FOR EACH ROW
    DECLARE 
        new_aid INT;
    BEGIN
        INSERT INTO accounts (username) VALUES (:new.username)
        RETURNING aid INTO new_aid;
        INSERT INTO account_infos VALUES (new_aid, :new.first_name, :new.last_name, :new.email);
    END;

INSERT INTO account_details (username, first_name, last_name, email)
VALUES ('user 6', 'Emma', 'E', 'emma@mail.com');

SELECT * FROM accounts;
SELECT * FROM account_infos;


/********/
/*More Trigger Example*/
/*this example show that trigger can also be used to enforce constraints in data*/
/*although it should not be overused - most of the time we want to validate data*/
/*at the application level, not the database level*/
DROP TABLE products;

CREATE TABLE products (
    pid INT GENERATED BY DEFAULT ON NULL AS IDENTITY,
    pname VARCHAR(50) NOT NULL,
    price NUMBER CHECK (price >= 0),
    CONSTRAINT productpk PRIMARY KEY (pid)
);

INSERT INTO products (pname, price) VALUES ('Pen',2.5);
INSERT INTO products (pname, price) VALUES ('Pencil',3.5);
INSERT INTO products (pname, price) VALUES ('Sharpie',4);
INSERT INTO products (pname, price) VALUES ('Notebook',2);

SELECT * FROM products;

DROP TABLE orders;

CREATE TABLE orders(
    orid INT GENERATED BY DEFAULT ON NULL AS IDENTITY,
    pid INT NOT NULL,
    quantity INT NOT NULL,
    total NUMBER NOT NULL,
    promotion NUMBER NOT NULL,
    datetime TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT orderfk FOREIGN KEY (pid) REFERENCES products (pid)
);

/*this trigger checks if the payment and promotion of an order 
match the price and quantity of item in orders*/
CREATE OR REPLACE TRIGGER insert_orders
BEFORE INSERT ON orders
FOR EACH ROW
    DECLARE
    item_price NUMBER;
    BEGIN
        SELECT price INTO item_price FROM products WHERE pid = :new.pid;
        IF :new.total + :new.promotion != item_price * :new.quantity THEN
                raise_application_error (-20999, 'total does not match price');
          else
                dbms_output.put_line('OK');
          end if;  
    END;

SHOW ERRORS;

INSERT INTO orders (pid,quantity,total,promotion) VALUES (1,4,10,0);
INSERT INTO orders (pid,quantity,total,promotion) VALUES (2,4,10,4);
INSERT INTO orders (pid,quantity,total,promotion) VALUES (1,4,5,0);

SELECT * FROM orders;

