/*****************************/
/*****Containers and PDBs*****/
/*****************************/
/*getting the current container name*/
SELECT SYS_CONTEXT('USERENV','CON_NAME') FROM dual;

/*getting the available pluggable databases and their statuses*/
SELECT PDB_ID, PDB_NAME, STATUS FROM DBA_PDBS ORDER BY PDB_ID;

/*checking if PDBs are open or not*/
/*OPEN_MODE being MOUNTED means the PDB is not opened*/
SELECT NAME, OPEN_MODE, RESTRICTED, OPEN_TIME FROM V$PDBS;


/*changing container in the current session*/
ALTER SESSION SET CONTAINER = samplescdb;
/*query some information from the samplescdb PDB*/
SELECT COUNT(*) FROM OE.CUSTOMERS;


/*move back to the root container*/
ALTER SESSION SET CONTAINER = cdb$root;



/****************************/
/*****Working with Users*****/
/****************************/

/****common user****/
/*we will create a user named c##admin with password 'abc123'*/
/*the c## prefix means this user is a common user*/
CREATE USER c##admin IDENTIFIED BY abc123 CONTAINER = ALL;

/*c##admin cannot connect or create a new session yet*/
/*we have to grant the required privileges*/
/*one way is to grant c##admin the role CONNECT*/

/*check which privileges that CONNECT has*/
SELECT * FROM dba_sys_privs WHERE grantee='CONNECT';

/*grant CONNECT role to c##admin*/
/*include CONTAINER = ALL so that c##admin can change container*/
GRANT CONNECT TO c##admin CONTAINER = ALL;


/*dropping user c##admin*/
/*we have to be in the root container to drop a common user*/
DROP USER c##admin;



/****local user****/
/*a local user cannot be created in the root container*/
CREATE USER user1 IDENTIFIED BY abc123; /*will yield errors*/

/*we must switch to a PDB, for example samplescdb*/
/*all scripts from this point will be running in the samplescdb container*/
/*DROP USER user1;*/
ALTER SESSION SET CONTAINER = samplescdb;
CREATE USER user1 IDENTIFIED BY abc123;

/*again, we have to grant connect permission*/
GRANT CONNECT TO user1;

/*we can check all the users in the database*/
/*and c##admin and user1 should be there*/
SELECT * FROM DBA_USERS;

/*user1 cannot create new tables since they do not have a quota,*/
/*tablespaces, and permission*/
/*let's create a new user with permission to create table*/
CREATE USER steve
      IDENTIFIED BY abc123
      DEFAULT TABLESPACE users
      TEMPORARY TABLESPACE temp
      QUOTA 15M ON users
      PROFILE default;

GRANT CONNECT TO steve;
GRANT CREATE TABLE to steve;


/*dropping user*/
/*if the user's schema does not have any object*/
DROP USER steve;

/*if the user's schema is not empty*/
DROP USER steve CASCADE;

/*when we want to block user acess to the database*/
/*but still want to keep their schema and objects*/
ALTER USER steve ACCOUNT LOCK;

/*or*/

REVOKE CONNECT, CREATE SESSION FROM steve;

/*but we can still access steve's table*/
/*assuming steve created a teststeve table*/
SELECT * FROM steve.teststeve;


/****************/
/***privileges***/
/****************/
/*grant view permission*/
/*we can also grant permission to INSERT, UPDATE, or DELETE if needed*/
GRANT SELECT ON oe.orders TO user1;

/*revoke permission*/
REVOKE SELECT ON oe.orders FROM user1;


/***********/
/***roles***/
/***********/
/*create role, assign privileges to role, then assign role to users*/
/*similar to before, we can assign permission to INSERT, UPDATE, and DELETE if needed*/
CREATE ROLE test_role1;

GRANT SELECT ON oe.orders TO test_role1;

GRANT test_role1 TO user1;

/*role with password*/
CREATE ROLE test_role2 IDENTIFIED BY abc123##;

GRANT SELECT ON oe.orders TO test_role2;
GRANT SELECT ON oe.customers TO test_role2;
GRANT SELECT ON oe.products TO test_role2;

GRANT test_role2 TO user1;
COMMIT;

/*user1 first have to change their role to test_role2*/
/*SET ROLE test_role2 IDENTIFIED BY abc123##; */
/*then they can select from the three tables, examples*/
/*SELECT COUNT(*) FROM oe.products; */

/**use view to limit access to certain tables**/
/*sample table with sensitive information*/

CREATE TABLE employees (
    eid NUMBER PRIMARY KEY,
    first_name VARCHAR(10),
    last_name VARCHAR(10),
    SSN VARCHAR(9),
    salary NUMBER
);

INSERT INTO employees VALUES (1,'Alice','Smith','123239434',80000);
INSERT INTO employees VALUES (2,'Bob','vander','164286366',70000);

/*first we create a view for employees with access limitted to only three attributes*/
CREATE VIEW employee_limit AS
SELECT eid, first_name, last_name
FROM employees;

/*then we grant user11 select permission to the view*/
/*now user1 only have access three attributes from the table*/
GRANT SELECT on employee_limit TO user1;
COMMIT;

/*user1 can then run this in their session*/
/*SELECT COUNT(*) FROM employee_limit;*/


/**use procedure to limit access to tables**/
/*revoke the previous select permission to test the procedure*/
REVOKE SELECT ON employee_limit FROM user1;

/*create procedure that query non-sensitive information*/
CREATE OR REPLACE PROCEDURE query_emp(e_id int )
AS
    /*we need a cursor here to run a regular select in the procedure*/
    cur sys_refcursor;
BEGIN
    OPEN cur FOR SELECT eid, first_name, last_name 
    FROM employees
    WHERE eid = e_id;
    DBMS_SQL.RETURN_RESULT(cur);
END;

/*then grant execute permission to user1*/
GRANT EXECUTE ON system.query_emp TO user1;
COMMIT;

/*now user1 can run this in their session*/
/* EXECUTE system.query_emp(1);*/